import dataclasses
import inspect
import os
import pathlib
import typing

from fbrp import registrar


@dataclasses.dataclass
class ProcDef:
    name: str
    root: pathlib.Path
    rule_file: pathlib.Path
    runtime: "BaseRuntime"
    cfg: dict
    deps: typing.List[str]
    env: dict


def process(
    name: str,
    root: str = None,
    runtime: "BaseRuntime" = None,
    cfg: dict = {},
    deps: typing.List[str] = [],
    env: dict = {},
):
    if name in registrar.defined_processes:
        raise ValueError(f"fbrp.process({name=}) defined multiple times.")

    rule_file = os.path.abspath(inspect.stack()[1].filename)
    rule_dir = os.path.dirname(rule_file)

    if root:
        root = os.path.normpath(os.path.join(rule_dir, root))
    else:
        root = rule_dir

    registrar.defined_processes[name] = ProcDef(
        name=name,
        root=root,
        runtime=runtime,
        cfg=cfg,
        deps=deps,
        rule_file=rule_file,
        env=env,
    )



from fbrp import life_cycle
from fbrp import registrar
import a0
import argparse
import json


@registrar.register_command("ps")
class ps_cmd:
    @classmethod
    def define_argparse(cls, parser: argparse.ArgumentParser):
        pass

    @staticmethod
    def exec(args: argparse.Namespace):
        state = life_cycle.system_state()
        if not state.procs:
            print("No processes found.")
            return

        name_col_width = max(len(name) for name in state.procs)

        for name, info in state.procs.items():
            suffix = ""
            if info.state == life_cycle.State.STOPPED:
                suffix = f"(stopped code={info.return_code})"
            print(name, " " * (name_col_width - len(name)), suffix)



from fbrp import life_cycle
from fbrp import registrar
from fbrp import util
from fbrp.cmd.base import BaseCommand
import a0
import argparse
import asyncio
import json
import os
import sys
import threading
import types
import typing


def transitive_closure(proc_names):
    all_proc_names = set()
    fringe = list(proc_names)
    while fringe:
        proc_name = fringe.pop()
        if proc_name in all_proc_names:
            continue
        try:
            fringe.extend(registrar.defined_processes[proc_name].deps)
        except KeyError:
            util.fail(f"Unknown process: {proc_name}")
        all_proc_names.add(proc_name)
    return all_proc_names


def get_proc_names(proc_names, include_deps):
    if not proc_names:
        return registrar.defined_processes.keys()

    if include_deps:
        proc_names = transitive_closure(proc_names)
    unknown_proc_names = [
        proc_name
        for proc_name in proc_names
        if proc_name not in registrar.defined_processes
    ]
    if unknown_proc_names:
        util.fail(f"Unknown proc_names: {', '.join(unknown_proc_names)}")
    if not proc_names:
        util.fail(f"No proc_names found")
    return proc_names


def down_existing(args: argparse.Namespace, names: typing.List[str]):
    def find_active_proc(system_state):
        return [
            name
            for name in names
            if name in system_state.procs
            and system_state.procs[name].state != life_cycle.State.STOPPED
        ]

    active_proc = find_active_proc(life_cycle.system_state())
    if not active_proc:
        return

    if not args.force:
        util.fail(f"Conflicting processes already running: {', '.join(active_proc)}")

    for name in active_proc:
        life_cycle.set_ask(name, life_cycle.Ask.DOWN)

    ns = types.SimpleNamespace()
    ns.cv = threading.Condition()
    ns.sat = False

    def callback(system_state):
        if not find_active_proc(system_state):
            with ns.cv:
                ns.sat = True
                ns.cv.notify()

    watcher = life_cycle.system_state_watcher(callback)

    with ns.cv:
        success = ns.cv.wait_for(lambda: ns.sat, timeout=3.0)

    if not success:
        util.fail(f"Existing processes did not down in a timely manner.")


@registrar.register_command("up")
class up_cmd(BaseCommand):
    @classmethod
    def define_argparse(cls, parser: argparse.ArgumentParser):
        parser.add_argument("--deps", default=True, action="store_true")
        parser.add_argument("--nodeps", dest="deps", action="store_false")
        parser.add_argument("--build", default=True, action="store_true")
        parser.add_argument("--nobuild", dest="build", action="store_false")
        parser.add_argument("--run", default=True, action="store_true")
        parser.add_argument("--norun", dest="run", action="store_false")
        parser.add_argument("-f", "--force", default=False, action="store_true")
        parser.add_argument("proc", action="append", nargs="*")

    @staticmethod
    def exec(args: argparse.Namespace):
        names = get_proc_names(args.proc[0], args.deps)
        names = [name for name in names if registrar.defined_processes[name].runtime]
        if not names:
            util.fail(f"No processes found")

        down_existing(args, names)

        if args.build:
            for name in names:
                proc_def = registrar.defined_processes[name]
                print(f"building {name}...")
                proc_def.runtime._build(name, proc_def, args)
                print(f"built {name}\n")

        if args.run:
            for name in names:
                print(f"running {name}...")
                life_cycle.set_ask(name, life_cycle.Ask.UP)

                if os.fork() != 0:
                    continue

                os.chdir("/")
                os.setsid()
                os.umask(0)

                if os.fork() != 0:
                    sys.exit(0)

                proc_def = registrar.defined_processes[name]

                # Set up configuration.
                with util.common_env_context(proc_def):
                    a0.Cfg(a0.env.topic()).write(json.dumps(proc_def.cfg))
                    life_cycle.set_launcher_running(name, True)
                    asyncio.run(proc_def.runtime._launcher(name, proc_def, args).run())
                    life_cycle.set_launcher_running(name, False)
                    sys.exit(0)



from fbrp import life_cycle
from fbrp import util
from fbrp.runtime.base import BaseLauncher, BaseRuntime
from fbrp.process import ProcDef
import a0
import aiodocker
import argparse
import asyncio
import contextlib
import docker
import io
import json
import os
import pwd
import sys
import typing


class Launcher(BaseLauncher):
    def __init__(
        self,
        image,
        mount: typing.List[str],
        name: str,
        proc_def: ProcDef,
        args: argparse.Namespace,
        **kwargs,
    ):
        self.image = image
        self.mount = mount
        self.name = name
        self.proc_def = proc_def
        self.args = args
        self.kwargs = kwargs

    async def run(self):
        life_cycle.set_state(self.name, life_cycle.State.STARTING)
        docker = aiodocker.Docker()

        container = f"fbrp_{self.name}"

        # Environmental variables.
        env = util.common_env(self.proc_def)
        for kv in self.kwargs.pop("Env", []):
            if "=" in kv:
                k, v = kv.split("=", 1)
                env[k] = v
            else:
                del env[kv]
        env.update(self.proc_def.env)

        # Docker labels.
        labels = {
            "fbrp.name": self.name,
            "fbrp.container": container,
        }
        labels.update(self.kwargs.pop("labels", {}))

        # Mount volumes.
        for f in [
            "group",
            "gshadow",
            "inputrc",
            "localtime",
            "passwd",
            "shadow",
            "subgid",
            "subuid",
            "sudoers",
        ]:
            self.mount.append(f"/etc/{f}:/etc/{f}:ro")

        self.mount.append("/var/lib/sss:/var/lib/sss:ro")

        id_info = pwd.getpwuid(os.getuid())

        # Compose all the arguments.
        run_kwargs = {
            "Image": self.image,
            "Env": ["=".join(kv) for kv in env.items()],
            "Labels": labels,
            "User": f"{id_info.pw_uid}:{id_info.pw_gid}",
            "HostConfig": {
                "Privileged": True,
                "NetworkMode": "host",
                "PidMode": "host",
                "IpcMode": "host",
                "Binds": self.mount,
                "GroupAdd": [str(i) for i in os.getgroups()],
            },
        }
        util.nested_dict_update(run_kwargs, self.kwargs)

        self.proc = await docker.containers.create_or_replace(container, run_kwargs)
        await self.proc.start()

        proc_info = await self.proc.show()
        self.proc_pid = proc_info["State"]["Pid"]
        life_cycle.set_state(self.name, life_cycle.State.STARTED)

        async def log_pipe(logger, pipe):
            async for line in pipe:
                logger(line)

        await asyncio.gather(
            log_pipe(util.stdout_logger(), self.proc.log(stdout=True, follow=True)),
            log_pipe(util.stderr_logger(), self.proc.log(stderr=True, follow=True)),
            self.log_psutil(),
            self.death_handler(),
            self.down_watcher(self.handle_down),
        )
        await docker.close()

    def get_pid(self):
        return self.proc_pid

    async def death_handler(self):
        result = await self.proc.wait()
        life_cycle.set_state(
            self.name, life_cycle.State.STOPPED, return_code=result["StatusCode"]
        )
        # TODO(lshamis): Restart policy goes here.

    async def handle_down(self):
        life_cycle.set_state(self.name, life_cycle.State.STOPPING)
        await self.proc.stop()
        with contextlib.suppress(asyncio.TimeoutError):
            await self.proc.wait(timeout=3.0)
        await self.proc.delete(force=True)


class Docker(BaseRuntime):
    def __init__(self, image=None, dockerfile=None, mount=[], build_kwargs={}, run_kwargs={}):
        if bool(image) == bool(dockerfile):
            util.fail("Docker process must define exactly one of image or dockerfile")
        self.image = image
        self.dockerfile = dockerfile
        self.mount = mount
        self.build_kwargs = build_kwargs
        self.run_kwargs = run_kwargs

    def _build(self, name: str, proc_def: ProcDef, args: argparse.Namespace):
        docker_api = docker.from_env()
        docker_api.lowlevel = docker.APIClient()

        if self.dockerfile:
            self.image = f"fbrp/netext:{util.random_string()}"

            dockerfile_path = os.path.join(
                os.path.dirname(proc_def.rule_file), self.dockerfile
            )

            build_kwargs = {
                "tag": self.image,
                "path": proc_def.root,
                "dockerfile": dockerfile_path,
                "rm": True,
            }
            build_kwargs.update(self.build_kwargs)

            for line in docker_api.lowlevel.build(**build_kwargs):
                lineinfo = json.loads(line.decode())
                if args.verbose and "stream" in lineinfo:
                    print(lineinfo["stream"].strip())
                elif "errorDetail" in lineinfo:
                    util.fail(json.dumps(lineinfo["errorDetail"], indent=2))

        uses_ldap = util.is_ldap_user()

        mount_map = {}
        for mnt in self.mount:
            parts = mnt.split(":")
            if len(parts) == 3:
                mount_map[parts[0]] = (parts[1], parts[2])
            elif len(parts) == 2:
                mount_map[parts[0]] = (parts[1], "")
            else:
                util.fail(f"Invalid mount: {mnt}")

        nfs_mounts = []
        for host, (container, _) in mount_map.items():
            try:
                os.makedirs(host)
            except FileExistsError:
                pass
            except FileNotFoundError:
                pass

            nfs_root = util.nfs_root(host)
            if nfs_root and nfs_root != host:
                nfs_mounts.append(
                    {
                        "host_nfs_root": nfs_root,
                        "container_nfs_root": f"/fbrp_nfs/{util.random_string()}",
                        "host": host,
                        "container": container,
                    }
                )

        for nfs_mount in nfs_mounts:
            if nfs_mount["host"] in mount_map:
                del mount_map[nfs_mount["host"]]

        if uses_ldap or nfs_mounts:
            dockerfile = [f"FROM {self.image}"]
            if uses_ldap:
                dockerfile.append(
                    " && ".join(
                        [
                            "RUN apt update",
                            "DEBIAN_FRONTEND=noninteractive apt install -y sssd",
                            "rm -rf /var/lib/apt/lists/*",
                        ]
                    )
                )
            for nfs_mount in nfs_mounts:
                relpath = os.path.relpath(nfs_mount["host"], nfs_mount["host_nfs_root"])
                container_fullpath = os.path.join(
                    nfs_mount["container_nfs_root"], relpath
                )
                mkdir = f"mkdir -p {os.path.dirname(nfs_mount['container'])}"
                link = f"ln -s {container_fullpath} {nfs_mount['container']}"
                dockerfile.append(f"RUN {mkdir} && {link}")

                mount_map[nfs_mount["host_nfs_root"]] = (
                    nfs_mount["container_nfs_root"],
                    "",
                )

            self.image = f"fbrp/netext:{util.random_string()}"

            for line in docker_api.lowlevel.build(
                fileobj=io.BytesIO("\n".join(dockerfile).encode("utf-8")),
                rm=True,
                tag=self.image,
                labels={
                    "fbrp.type": "netext",
                    "fbrp.tmp": "1",
                },
            ):
                lineinfo = json.loads(line.decode())
                if args.verbose and "stream" in lineinfo:
                    print(lineinfo["stream"].strip())
                elif "errorDetail" in lineinfo:
                    util.fail(json.dumps(lineinfo["errorDetail"], indent=2))

        self.mount = [
            f"{host_path}:{container_path}:{options}"
            for host_path, (container_path, options) in mount_map.items()
        ]

    def _launcher(self, name: str, proc_def: ProcDef, args: argparse.Namespace):
        return Launcher(self.image, self.mount, name, proc_def, args, **self.run_kwargs)



from gc import callbacks
import a0
from fbrp import life_cycle
from fbrp import registrar
import argparse
import threading
import types


@registrar.register_command("wait")
class wait_cmd:
    @classmethod
    def define_argparse(cls, parser: argparse.ArgumentParser):
        parser.add_argument("proc", action="append", nargs="*")
        parser.add_argument("-t", "--timeout", action="store", type=float, default=0)

    @staticmethod
    def exec(args: argparse.Namespace):
        procs = life_cycle.system_state().procs.keys()

        given_proc_names = args.proc[0]
        if given_proc_names:
            procs = set(procs) & set(given_proc_names)

        ns = types.SimpleNamespace()
        ns.cv = threading.Condition()
        ns.sat = False

        def callback(sys_state):
            for proc, info in sys_state.procs.items():
                if proc in procs and info.state != life_cycle.State.STOPPED:
                    return
            with ns.cv:
                ns.sat = True
                ns.cv.notify()

        watcher = life_cycle.system_state_watcher(callback)

        with ns.cv:
            ns.cv.wait_for(lambda: ns.sat, timeout=args.timeout or None)



import argparse


class BaseCommand:
    @classmethod
    def define_argparse(cls, parser: argparse.ArgumentParser):
        raise NotImplementedError("Command hasn't implemented define_argparse!")

    @staticmethod
    def exec(args: argparse.Namespace):
        raise NotImplementedError("Command hasn't implemented exec!")



from fbrp.process import ProcDef
import typing

defined_commands: typing.Dict[str, "BaseCommand"] = {}
defined_processes: typing.Dict[str, ProcDef] = {}


def register_command(name):
    def wrap(cls):
        defined_commands[name] = cls
        return cls

    return wrap



import dataclasses
from fbrp import life_cycle
from fbrp import util
from fbrp.runtime.base import BaseLauncher, BaseRuntime
from fbrp.process import ProcDef
import asyncio
import dataclasses
import json
import os
import shlex
import shutil
import signal
import subprocess
import argparse
import typing
import yaml as pyyaml


@dataclasses.dataclass
class CondaEnv:
    channels: typing.List[str]
    dependencies: typing.List[typing.Union[str, dict]]

    @staticmethod
    def load(flo):
        result = pyyaml.safe_load(flo)
        return CondaEnv(result.get("channels", []), result.get("dependencies", []))

    @staticmethod
    def from_env(name):
        return CondaEnv.load(
            subprocess.run(
                ["conda", "env", "export", "-n", name], check=True, capture_output=True
            ).stdout
        )

    @staticmethod
    def merge(lhs: "CondaEnv", rhs: "CondaEnv") -> "CondaEnv":
        channels = sorted(set(lhs.channels + rhs.channels))

        deps = []
        pip_deps = []
        for dep in lhs.dependencies + rhs.dependencies:
            if type(dep) == dict and dep.get("pip"):
                pip_deps.extend(dep["pip"])
            else:
                deps.append(dep)

        deps = sorted(set(deps))

        if pip_deps:
            pip_deps = sorted(set(pip_deps))
            deps.append({"pip": pip_deps})

        return CondaEnv(channels, deps)

    def fix_pip(self):
        if any(dep is str and dep.startswith("pip") for dep in self.dependencies):
            return
        for dep in self.dependencies:
            if type(dep) == dict and dep.get("pip"):
                self.dependencies.append("pip")
                return


class Launcher(BaseLauncher):
    def __init__(
        self,
        run_command: typing.List[str],
        name: str,
        proc_def: ProcDef,
        args: argparse.Namespace,
    ):
        self.run_command = run_command
        self.name = name
        self.proc_def = proc_def
        self.args = args

    async def activate_conda_env(self) -> dict:
        # We grab the conda env variables separate from executing the run
        # command to simplify detecting pid and removing some race conditions.
        subprocess_env = os.environ.copy()
        subprocess_env.update(self.proc_def.env)
        envinfo = await asyncio.create_subprocess_shell(
            f"""
                eval "$(conda shell.bash hook)"
                conda activate fbrp_{self.name}
                cp -f /proc/self/environ /tmp/fbrp_conda_{self.name}.env
            """,
            stdout=asyncio.subprocess.PIPE,
            executable="/bin/bash",
            cwd=self.proc_def.root,
            env=subprocess_env,
        )

        await envinfo.wait()
        lines = open(f"/tmp/fbrp_conda_{self.name}.env").read().split("\0")
        conda_env = dict(line.split("=", 1) for line in lines if "=" in line)
        return conda_env

    async def run_cmd_in_env(self, conda_env):
        self.proc = await asyncio.create_subprocess_shell(
            util.shell_join(self.run_command),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            executable="/bin/bash",
            cwd=self.proc_def.root,
            env=conda_env,
        )

    async def gather_cmd_outputs(self):
        async def log_pipe(logger, pipe):
            while True:
                line = await pipe.readline()
                if line:
                    logger(line)
                else:
                    break

        await asyncio.gather(
            log_pipe(util.stdout_logger(), self.proc.stdout),
            log_pipe(util.stderr_logger(), self.proc.stderr),
            self.log_psutil(),
            self.death_handler(),
            self.down_watcher(self.handle_down),
        )

    async def run(self):
        life_cycle.set_state(self.name, life_cycle.State.STARTING)
        conda_env = await self.activate_conda_env()
        await self.run_cmd_in_env(conda_env)
        life_cycle.set_state(self.name, life_cycle.State.STARTED)
        await self.gather_cmd_outputs()

    def get_pid(self):
        return self.proc.pid

    async def exit_cmd_in_env(self) -> int:
        await self.proc.wait()
        return self.proc.returncode

    async def death_handler(self):
        ret_code = await self.exit_cmd_in_env()
        life_cycle.set_state(self.name, life_cycle.State.STOPPED, return_code=ret_code)
        # TODO(lshamis): Restart policy goes here.

    async def handle_down(self):
        try:
            if self.proc.returncode is not None:
                return
            proc_pid = self.get_pid()

            life_cycle.set_state(self.name, life_cycle.State.STOPPING)
            os.kill(proc_pid, signal.SIGTERM)

            for _ in range(100):
                if not os.kill(proc_pid, 0):
                    break
                await asyncio.sleep(0.03)

            if os.kill(proc_pid, 0):
                os.kill(proc_pid, signal.SIGKILL)
        except:
            pass


class Conda(BaseRuntime):
    def __init__(
        self,
        run_command,
        yaml=None,
        env=None,
        channels=[],
        dependencies=[],
        setup_commands=[],
        use_mamba=None,
    ):
        self.yaml = yaml
        self.env = env
        self.run_command = run_command
        self.conda_env = CondaEnv(channels[:], dependencies[:])
        self.setup_commands = setup_commands
        self.use_mamba = (
            use_mamba if use_mamba is not None else bool(shutil.which("mamba"))
        )

    def _build(self, name: str, proc_def: ProcDef, args: argparse.Namespace):
        if self.yaml:
            yaml_path = os.path.join(proc_def.root, self.yaml)
            self.conda_env = CondaEnv.merge(
                self.conda_env, CondaEnv.load(open(yaml_path, "r"))
            )

        if self.env:
            self.conda_env = CondaEnv.merge(self.conda_env, CondaEnv.from_env(self.env))

        self.conda_env.fix_pip()

        env_name = f"fbrp_{name}"
        env_path = f"/tmp/fbrp_conda_{name}.yml"

        with open(env_path, "w") as env_fp:
            json.dump(
                {
                    "name": env_name,
                    "channels": self.conda_env.channels,
                    "dependencies": self.conda_env.dependencies,
                },
                env_fp,
                indent=2,
            )

        print(f"creating conda env for {name}. This will take a minute...")

        update_bin = "mamba" if self.use_mamba else "conda"
        # https://github.com/conda/conda/issues/7279
        # Updating an existing environment does not remove old packages, even with --prune.
        subprocess.run(
            [update_bin, "env", "remove", "-n", env_name],
            capture_output=not args.verbose,
        )
        result = subprocess.run(
            [update_bin, "env", "update", "--prune", "-f", env_path],
            capture_output=not args.verbose,
        )
        if result.returncode:
            util.fail(f"Failed to set up conda env: {result.stderr}")

        if self.setup_commands:
            print(f"setting up conda env for {name}")
            setup_command = "\n".join(
                [util.shell_join(cmd) for cmd in self.setup_commands]
            )
            result = subprocess.run(
                f"""
                    eval "$(conda shell.bash hook)"
                    conda activate {env_name}
                    cd {proc_def.root}
                    {setup_command}
                """,
                shell=True,
                executable="/bin/bash",
                capture_output=not args.verbose,
            )
            if result.returncode:
                util.fail(f"Failed to set up conda env: {result.stderr}")

    def _launcher(self, name: str, proc_def: ProcDef, args: argparse.Namespace):
        return Launcher(self.run_command, name, proc_def, args)


__all__ = ["Conda"]



from fbrp import registrar
from fbrp.process import process
from fbrp.runtime.conda import Conda
from fbrp.runtime.docker import Docker
from fbrp.util import NoEscape
import argparse


def main():
    import fbrp.cmd.down
    import fbrp.cmd.list
    import fbrp.cmd.logs
    import fbrp.cmd.ps
    import fbrp.cmd.up
    import fbrp.cmd.wait

    parser = argparse.ArgumentParser(prog="fbrp")
    parser.add_argument("-v", "--verbose", default=False, action="store_true")
    subparsers = parser.add_subparsers()
    subparsers.required = True
    subparsers.dest = "cmd"

    for cmd_name, cmd_cls in registrar.defined_commands.items():
        cmd_parser = subparsers.add_parser(cmd_name)
        cmd_cls.define_argparse(cmd_parser)
        cmd_parser.set_defaults(func=cmd_cls.exec)

    args = parser.parse_args()

    args.func(args)


__all__ = ["main", "process", "NoEscape", "Docker", "Conda"]



from fbrp import life_cycle
from fbrp import registrar
import argparse


@registrar.register_command("down")
class down_cmd:
    @classmethod
    def define_argparse(cls, parser: argparse.ArgumentParser):
        parser.add_argument("proc", action="append", nargs="*")

    @staticmethod
    def exec(args: argparse.Namespace):
        procs = life_cycle.system_state().procs.keys()

        given_proc_names = args.proc[0]
        if given_proc_names:
            procs = set(procs) & set(given_proc_names)

        for proc_name in procs:
            life_cycle.set_ask(proc_name, life_cycle.Ask.DOWN)



from fbrp import registrar
import argparse


@registrar.register_command("list")
class list_cmd:
    @classmethod
    def define_argparse(cls, parser: argparse.ArgumentParser):
        parser.add_argument("proc", action="append", nargs="*")

    @staticmethod
    def exec(args: argparse.Namespace):
        print("Defined Processes:")
        for proc in registrar.defined_processes:
            print(f"  {proc}")



import asyncio
from fbrp import life_cycle
from fbrp.process import ProcDef
import a0
import argparse
import asyncio
import contextlib
import psutil
import json


class BaseLauncher:
    def __init__(self):
        pass

    async def run(self, name: str, proc_def: ProcDef, args: argparse.Namespace):
        raise NotImplementedError("Launcher hasn't implemented run!")

    def get_pid(self):
        raise NotImplementedError("Launcher hasn't implemented get_pid!")

    async def down_watcher(self, ondown):
        async for proc_info in life_cycle.aio_proc_info_watcher(self.name):
            if proc_info.ask == life_cycle.Ask.DOWN:
                await ondown()
                break

    async def log_psutil(self):
        down_requested_event = asyncio.Event()

        async def ondown():
            down_requested_event.set()

        asyncio.ensure_future(self.down_watcher(ondown))

        out = a0.Publisher(f"fbrp/psutil/{self.name}")
        while True:
            with contextlib.suppress(asyncio.TimeoutError):
                # TODO(lshamis): Make polling interval configurable.
                await asyncio.wait_for(down_requested_event.wait(), 1.0)
            if down_requested_event.is_set():
                break
            pid = self.get_pid()
            if not pid:
                # Note: Likely being restarted.
                continue

            try:
                proc = psutil.Process(pid)
                out.pub(json.dumps(proc.as_dict()))
            except psutil.NoSuchProcess:
                pass


class BaseRuntime:
    def __init__(self):
        pass

    def _build(self, name: str, proc_def: ProcDef, args: argparse.Namespace):
        raise NotImplementedError("Runtime hasn't implemented build!")

    def _launcher(
        self, name: str, proc_def: ProcDef, args: argparse.Namespace
    ) -> BaseLauncher:
        raise NotImplementedError("Runtime hasn't implemented a launcher!")



from fbrp import util
from fbrp import registrar
import a0
import argparse
import random
import signal
import sys


@registrar.register_command("logs")
class logs_cmd:
    @classmethod
    def define_argparse(cls, parser: argparse.ArgumentParser):
        parser.add_argument("--old", default=False, action="store_true")
        parser.add_argument("proc", action="append", nargs="*")

    @staticmethod
    def exec(args: argparse.Namespace):
        procs = {
            name: def_
            for name, def_ in registrar.defined_processes.items()
            if def_.runtime
        }

        given_proc_names = args.proc[0]
        if given_proc_names:
            procs = {
                name: def_ for name, def_ in procs.items() if name in given_proc_names
            }

        if not procs:
            util.fail(f"No processes found to log")

        colors = [
            "\u001b[31m",  # "Red"
            "\u001b[32m",  # "Green"
            "\u001b[33m",  # "Yellow"
            "\u001b[34m",  # "Blue"
            "\u001b[35m",  # "Magenta"
            "\u001b[36m",  # "Cyan"
        ]
        random.shuffle(colors)
        reset_color = "\u001b[0m"

        width = max(len(name) for name in procs)

        log_listeners = []

        def make_listener(i, name, def_):
            prefix = f"{colors[i % len(colors)]}{name}" + " " * (width - len(name))
            msg_tmpl = f"{prefix} | {{msg}}{reset_color}"

            def callback(pkt):
                print(msg_tmpl.format(msg=pkt.payload))

            with util.common_env_context(def_):
                log_listeners.append(
                    a0.LogListener(
                        name,
                        a0.LogLevel.DBG,
                        a0.INIT_OLDEST if args.old else a0.INIT_AWAIT_NEW,
                        a0.ITER_NEXT,
                        callback,
                    )
                )

        for i, (name, def_) in enumerate(procs.items()):
            make_listener(i, name, def_)

        def onsignal(signum, frame):
            sys.exit(0)

        signal.signal(signal.SIGINT, onsignal)
        signal.signal(signal.SIGTERM, onsignal)
        signal.pause()
